@@ -10,12 +10,24 @@ from functools import lru_cache, partial
 from keyword import iskeyword
 
 import attr
-
 import pycodestyle
 
-__version__ = "21.9.2"
+__version__ = "22.3.20"
 
 LOG = logging.getLogger("flake8.bugbear")
+CONTEXTFUL_NODES = (
+    ast.Module,
+    ast.ClassDef,
+    ast.AsyncFunctionDef,
+    ast.FunctionDef,
+    ast.Lambda,
+    ast.ListComp,
+    ast.SetComp,
+    ast.DictComp,
+    ast.GeneratorExp,
+)
+
+Context = namedtuple("Context", ["node", "stack"])
 
 
 @attr.s(hash=False)
@@ -33,7 +45,17 @@ class BugBearChecker:
     def run(self):
         if not self.tree or not self.lines:
             self.load_file()
-        visitor = self.visitor(filename=self.filename, lines=self.lines)
+
+        if self.options and hasattr(self.options, "extend_immutable_calls"):
+            b008_extend_immutable_calls = set(self.options.extend_immutable_calls)
+        else:
+            b008_extend_immutable_calls = set()
+
+        visitor = self.visitor(
+            filename=self.filename,
+            lines=self.lines,
+            b008_extend_immutable_calls=b008_extend_immutable_calls,
+        )
         visitor.visit(self.tree)
         for e in itertools.chain(visitor.errors, self.gen_line_based_checks()):
             if self.should_warn(e.message[:4]):
@@ -45,8 +67,31 @@ class BugBearChecker:
         The following simple checks are based on the raw lines, not the AST.
         """
         for lineno, line in enumerate(self.lines, start=1):
+            # Special case: ignore long shebang (following pycodestyle).
+            if lineno == 1 and line.startswith("#!"):
+                continue
+
             length = len(line) - 1
-            if length > 1.1 * self.max_line_length:
+            if length > 1.1 * self.max_line_length and line.strip():
+                # Special case long URLS and paths to follow pycodestyle.
+                # Would use the `pycodestyle.maximum_line_length` directly but
+                # need to supply it arguments that are not available so chose
+                # to replicate instead.
+                chunks = line.split()
+
+                is_line_comment_url_path = len(chunks) == 2 and chunks[0] == "#"
+
+                just_long_url_path = len(chunks) == 1
+
+                num_leading_whitespaces = len(line) - len(chunks[-1])
+                too_many_leading_white_spaces = (
+                    num_leading_whitespaces >= self.max_line_length - 7
+                )
+
+                skip = is_line_comment_url_path or just_long_url_path
+                if skip and not too_many_leading_white_spaces:
+                    continue
+
                 yield B950(lineno, length, vars=(length, self.max_line_length))
 
     @classmethod
@@ -74,8 +119,15 @@ class BugBearChecker:
     def add_options(optmanager):
         """Informs flake8 to ignore B9xx by default."""
         optmanager.extend_default_ignore(disabled_by_default)
+        optmanager.add_option(
+            "--extend-immutable-calls",
+            comma_separated_list=True,
+            parse_from_config=True,
+            default=[],
+            help="Skip B008 test for additional immutable calls.",
+        )
 
-    @lru_cache()
+    @lru_cache()  # noqa: B019
     def should_warn(self, code):
         """Returns `True` if Bugbear should emit a particular warning.
 
@@ -86,12 +138,17 @@ class BugBearChecker:
 
         As documented in the README, the user is expected to explicitly select
         the warnings.
+
+        NOTE: This method is deprecated and will be removed in a future release. It is
+        recommended to use `extend-ignore` and `extend-select` in your flake8
+        configuration to avoid implicitly altering selected and ignored codes.
         """
         if code[:2] != "B9":
             # Normal warnings are safe for emission.
             return True
 
         if self.options is None:
+            # Without options configured, Bugbear will emit B9 but flake8 will ignore
             LOG.info(
                 "Options not provided to Bugbear, optional warning %s selected.", code
             )
@@ -101,6 +158,13 @@ class BugBearChecker:
             if code[:i] in self.options.select:
                 return True
 
+            # flake8 >=4.0: Also check for codes in extend_select
+            if (
+                hasattr(self.options, "extend_select")
+                and code[:i] in self.options.extend_select
+            ):
+                return True
+
         LOG.info(
             "Optional warning %s not present in selected warnings: %r. Not "
             "firing it at all.",
@@ -147,10 +211,12 @@ def _typesafe_issubclass(cls, class_or_tuple):
 class BugBearVisitor(ast.NodeVisitor):
     filename = attr.ib()
     lines = attr.ib()
+    b008_extend_immutable_calls = attr.ib(default=attr.Factory(set))
     node_stack = attr.ib(default=attr.Factory(list))
     node_window = attr.ib(default=attr.Factory(list))
     errors = attr.ib(default=attr.Factory(list))
     futures = attr.ib(default=attr.Factory(set))
+    contexts = attr.ib(default=attr.Factory(list))
 
     NODE_WINDOW_SIZE = 4
 
@@ -161,13 +227,30 @@ class BugBearVisitor(ast.NodeVisitor):
             print(name)
             return self.__getattribute__(name)
 
+    @property
+    def node_stack(self):
+        if len(self.contexts) == 0:
+            return []
+
+        context, stack = self.contexts[-1]
+        return stack
+
     def visit(self, node):
+        is_contextful = isinstance(node, CONTEXTFUL_NODES)
+
+        if is_contextful:
+            context = Context(node, [])
+            self.contexts.append(context)
+
         self.node_stack.append(node)
         self.node_window.append(node)
         self.node_window = self.node_window[-self.NODE_WINDOW_SIZE :]
         super().visit(node)
         self.node_stack.pop()
 
+        if is_contextful:
+            self.contexts.pop()
+
     def visit_ExceptHandler(self, node):
         if node.type is None:
             self.errors.append(
@@ -187,20 +270,19 @@ class BugBearVisitor(ast.NodeVisitor):
                 #    (MyError, BaseException)  # everything derives from the Base
                 #    (Exception, TypeError)  # builtins where one subclasses another
                 #    (IOError, OSError)  # IOError is an alias of OSError since Python3.3
-                # but note that other cases are impractical to hande from the AST.
+                # but note that other cases are impractical to handle from the AST.
                 # We expect this is mostly useful for users who do not have the
                 # builtin exception hierarchy memorised, and include a 'shadowed'
                 # subtype without realising that it's redundant.
                 good = sorted(set(names), key=names.index)
                 if "BaseException" in good:
                     good = ["BaseException"]
-                # Find and remove aliases exceptions and only leave the primary alone
-                primaries = filter(
-                    lambda primary: primary in good, B014.exception_aliases.keys()
-                )
-                for primary in primaries:
-                    aliases = B014.exception_aliases[primary]
-                    good = list(filter(lambda e: e not in aliases, good))
+                # Remove redundant exceptions that the automatic system either handles
+                # poorly (usually aliases) or can't be checked (e.g. it's not an
+                # built-in exception).
+                for primary, equivalents in B014.redundant_exceptions.items():
+                    if primary in good:
+                        good = [g for g in good if g not in equivalents]
 
                 for name, other in itertools.permutations(tuple(good), 2):
                     if _typesafe_issubclass(
@@ -233,21 +315,21 @@ class BugBearVisitor(ast.NodeVisitor):
             with suppress(AttributeError, IndexError):
                 if (
                     node.func.id in ("getattr", "hasattr")
-                    and node.args[1].s == "__call__"  # noqa: W503
+                    and node.args[1].s == "__call__"
                 ):
                     self.errors.append(B004(node.lineno, node.col_offset))
                 if (
                     node.func.id == "getattr"
-                    and len(node.args) == 2  # noqa: W503
-                    and _is_identifier(node.args[1])  # noqa: W503
-                    and not iskeyword(node.args[1].s)  # noqa: W503
+                    and len(node.args) == 2
+                    and _is_identifier(node.args[1])
+                    and not iskeyword(node.args[1].s)
                 ):
                     self.errors.append(B009(node.lineno, node.col_offset))
                 elif (
                     node.func.id == "setattr"
-                    and len(node.args) == 3  # noqa: W503
-                    and _is_identifier(node.args[1])  # noqa: W503
-                    and not iskeyword(node.args[1].s)  # noqa: W503
+                    and len(node.args) == 3
+                    and _is_identifier(node.args[1])
+                    and not iskeyword(node.args[1].s)
                 ):
                     self.errors.append(B010(node.lineno, node.col_offset))
 
@@ -263,6 +345,7 @@ class BugBearVisitor(ast.NodeVisitor):
 
     def visit_For(self, node):
         self.check_for_b007(node)
+        self.check_for_b020(node)
         self.generic_visit(node)
 
     def visit_Assert(self, node):
@@ -278,10 +361,15 @@ class BugBearVisitor(ast.NodeVisitor):
         self.check_for_b901(node)
         self.check_for_b902(node)
         self.check_for_b006(node)
+        self.check_for_b018(node)
+        self.check_for_b019(node)
+        self.check_for_b021(node)
         self.generic_visit(node)
 
     def visit_ClassDef(self, node):
         self.check_for_b903(node)
+        self.check_for_b018(node)
+        self.check_for_b021(node)
         self.generic_visit(node)
 
     def visit_Try(self, node):
@@ -299,12 +387,15 @@ class BugBearVisitor(ast.NodeVisitor):
 
     def visit_With(self, node):
         self.check_for_b017(node)
+        self.check_for_b022(node)
         self.generic_visit(node)
 
     def compose_call_path(self, node):
         if isinstance(node, ast.Attribute):
             yield from self.compose_call_path(node.value)
             yield node.attr
+        elif isinstance(node, ast.Call):
+            yield from self.compose_call_path(node.func)
         elif isinstance(node, ast.Name):
             yield node.id
 
@@ -338,7 +429,10 @@ class BugBearVisitor(ast.NodeVisitor):
                 call_path = ".".join(self.compose_call_path(default.func))
                 if call_path in B006.mutable_calls:
                     self.errors.append(B006(default.lineno, default.col_offset))
-                elif call_path not in B008.immutable_calls:
+                elif (
+                    call_path
+                    not in B008.immutable_calls | self.b008_extend_immutable_calls
+                ):
                     # Check if function call is actually a float infinity/NaN literal
                     if call_path == "float" and len(default.args) == 1:
                         float_arg = default.args[0]
@@ -422,15 +516,56 @@ class BugBearVisitor(ast.NodeVisitor):
         item_context = item.context_expr
         if (
             hasattr(item_context, "func")
-            and hasattr(item_context.func, "attr")  # noqa W503
-            and item_context.func.attr == "assertRaises"  # noqa W503
-            and len(item_context.args) == 1  # noqa W503
-            and isinstance(item_context.args[0], ast.Name)  # noqa W503
-            and item_context.args[0].id == "Exception"  # noqa W503
-            and not item.optional_vars  # noqa W503
+            and hasattr(item_context.func, "attr")
+            and item_context.func.attr == "assertRaises"
+            and len(item_context.args) == 1
+            and isinstance(item_context.args[0], ast.Name)
+            and item_context.args[0].id == "Exception"
+            and not item.optional_vars
         ):
             self.errors.append(B017(node.lineno, node.col_offset))
 
+    def check_for_b019(self, node):
+        if (
+            len(node.decorator_list) == 0
+            or len(self.contexts) < 2
+            or not isinstance(self.contexts[-2].node, ast.ClassDef)
+        ):
+            return
+
+        # Preserve decorator order so we can get the lineno from the decorator node
+        # rather than the function node (this location definition changes in Python 3.8)
+        resolved_decorators = (
+            ".".join(self.compose_call_path(decorator))
+            for decorator in node.decorator_list
+        )
+        for idx, decorator in enumerate(resolved_decorators):
+            if decorator in {"classmethod", "staticmethod"}:
+                return
+
+            if decorator in B019.caches:
+                self.errors.append(
+                    B019(
+                        node.decorator_list[idx].lineno,
+                        node.decorator_list[idx].col_offset,
+                    )
+                )
+                return
+
+    def check_for_b020(self, node):
+        targets = NameFinder()
+        targets.visit(node.target)
+        ctrl_names = set(targets.names)
+
+        iterset = NameFinder()
+        iterset.visit(node.iter)
+        iterset_names = set(iterset.names)
+
+        for name in sorted(ctrl_names):
+            if name in iterset_names:
+                n = targets.names[name][0]
+                self.errors.append(B020(n.lineno, n.col_offset, vars=(name,)))
+
     def check_for_b904(self, node):
         """Checks `raise` without `from` inside an `except` clause.
 
@@ -479,9 +614,12 @@ class BugBearVisitor(ast.NodeVisitor):
                 break
 
     def check_for_b902(self, node):
-        if not isinstance(self.node_stack[-2], ast.ClassDef):
+        if len(self.contexts) < 2 or not isinstance(
+            self.contexts[-2].node, ast.ClassDef
+        ):
             return
 
+        cls = self.contexts[-2].node
         decorators = NameFinder()
         decorators.visit(node.decorator_list)
 
@@ -490,11 +628,11 @@ class BugBearVisitor(ast.NodeVisitor):
             # `cls`?
             return
 
-        bases = {b.id for b in self.node_stack[-2].bases if isinstance(b, ast.Name)}
+        bases = {b.id for b in cls.bases if isinstance(b, ast.Name)}
         if "type" in bases:
             if (
                 "classmethod" in decorators.names
-                or node.name in B902.implicit_classmethods  # noqa: W503
+                or node.name in B902.implicit_classmethods
             ):
                 expected_first_args = B902.metacls
                 kind = "metaclass class"
@@ -504,7 +642,7 @@ class BugBearVisitor(ast.NodeVisitor):
         else:
             if (
                 "classmethod" in decorators.names
-                or node.name in B902.implicit_classmethods  # noqa: W503
+                or node.name in B902.implicit_classmethods
             ):
                 expected_first_args = B902.cls
                 kind = "class"
@@ -549,16 +687,16 @@ class BugBearVisitor(ast.NodeVisitor):
         body = node.body
         if (
             body
-            and isinstance(body[0], ast.Expr)  # noqa: W503
-            and isinstance(body[0].value, ast.Str)  # noqa: W503
+            and isinstance(body[0], ast.Expr)
+            and isinstance(body[0].value, ast.Str)
         ):
             # Ignore the docstring
             body = body[1:]
 
         if (
             len(body) != 1
-            or not isinstance(body[0], ast.FunctionDef)  # noqa: W503
-            or body[0].name != "__init__"  # noqa: W503
+            or not isinstance(body[0], ast.FunctionDef)
+            or body[0].name != "__init__"
         ):
             # only classes with *just* an __init__ method are interesting
             return
@@ -575,6 +713,47 @@ class BugBearVisitor(ast.NodeVisitor):
 
         self.errors.append(B903(node.lineno, node.col_offset))
 
+    def check_for_b018(self, node):
+        for subnode in node.body:
+            if not isinstance(subnode, ast.Expr):
+                continue
+            if isinstance(
+                subnode.value,
+                (
+                    ast.Num,
+                    ast.Bytes,
+                    ast.NameConstant,
+                    ast.List,
+                    ast.Set,
+                    ast.Dict,
+                ),
+            ):
+                self.errors.append(B018(subnode.lineno, subnode.col_offset))
+
+    def check_for_b021(self, node):
+        if (
+            node.body
+            and isinstance(node.body[0], ast.Expr)
+            and isinstance(node.body[0].value, ast.JoinedStr)
+        ):
+            self.errors.append(
+                B021(node.body[0].value.lineno, node.body[0].value.col_offset)
+            )
+
+    def check_for_b022(self, node):
+        item = node.items[0]
+        item_context = item.context_expr
+        if (
+            hasattr(item_context, "func")
+            and hasattr(item_context.func, "value")
+            and hasattr(item_context.func.value, "id")
+            and item_context.func.value.id == "contextlib"
+            and hasattr(item_context.func, "attr")
+            and item_context.func.attr == "suppress"
+            and len(item_context.args) == 0
+        ):
+            self.errors.append(B022(node.lineno, node.col_offset))
+
 
 @attr.s
 class NameFinder(ast.NodeVisitor):
@@ -736,15 +915,19 @@ B014 = Error(
         "Write `except {2}{1}:`, which catches exactly the same exceptions."
     )
 )
-B014.exception_aliases = {
+B014.redundant_exceptions = {
     "OSError": {
+        # All of these are actually aliases of OSError since Python 3.3
         "IOError",
         "EnvironmentError",
         "WindowsError",
         "mmap.error",
         "socket.error",
         "select.error",
-    }
+    },
+    "ValueError": {
+        "binascii.Error",
+    },
 }
 B015 = Error(
     message=(
@@ -767,6 +950,43 @@ B017 = Error(
         "context manager form of assertRaises."
     )
 )
+B018 = Error(
+    message=(
+        "B018 Found useless expression. Either assign it to a variable or remove it."
+    )
+)
+B019 = Error(
+    message=(
+        "B019 Use of `functools.lru_cache` or `functools.cache` on class methods "
+        "can lead to memory leaks. The cache may retain instance references, "
+        "preventing garbage collection."
+    )
+)
+B019.caches = {
+    "functools.cache",
+    "functools.lru_cache",
+    "cache",
+    "lru_cache",
+}
+B020 = Error(
+    message=(
+        "B020 Found for loop that reassigns the iterable it is iterating "
+        + "with each iterable value."
+    )
+)
+B021 = Error(
+    message=(
+        "B021 f-string used as docstring."
+        "This will be interpreted by python as a joined string rather than a docstring."
+    )
+)
+B022 = Error(
+    message=(
+        "B022 No arguments passed to `contextlib.suppress`."
+        "No exceptions will be suppressed and therefore this"
+        "context manager is redundant."
+    )
+)
 
 # Warnings disabled by default.
 B901 = Error(
