@@ -1,15 +1,15 @@
 import ast
 import os
-from pathlib import Path
 import site
 import subprocess
 import sys
 import unittest
+from argparse import Namespace
+from pathlib import Path
 
 from hypothesis import HealthCheck, given, settings
 from hypothesmith import from_grammar
 
-from bugbear import BugBearChecker, BugBearVisitor
 from bugbear import (
     B001,
     B002,
@@ -28,11 +28,18 @@ from bugbear import (
     B015,
     B016,
     B017,
-    B904,
+    B018,
+    B019,
+    B020,
+    B021,
+    B022,
     B901,
     B902,
     B903,
+    B904,
     B950,
+    BugBearChecker,
+    BugBearVisitor,
 )
 
 
@@ -63,7 +70,7 @@ class BugbearTestCase(unittest.TestCase):
         filename = Path(__file__).absolute().parent / "b003.py"
         bbc = BugBearChecker(filename=str(filename))
         errors = list(bbc.run())
-        self.assertEqual(errors, self.errors(B003(10, 0)))
+        self.assertEqual(errors, self.errors(B003(9, 0)))
 
     def test_b004(self):
         filename = Path(__file__).absolute().parent / "b004.py"
@@ -123,6 +130,21 @@ class BugbearTestCase(unittest.TestCase):
             ),
         )
 
+    def test_b008_extended(self):
+        filename = Path(__file__).absolute().parent / "b008_extended.py"
+
+        mock_options = Namespace(
+            extend_immutable_calls=["fastapi.Depends", "fastapi.Query"]
+        )
+
+        bbc = BugBearChecker(filename=str(filename), options=mock_options)
+        errors = list(bbc.run())
+
+        self.assertEqual(
+            errors,
+            self.errors(B008(15, 66)),
+        )
+
     def test_b009_b010(self):
         filename = Path(__file__).absolute().parent / "b009_b010.py"
         bbc = BugBearChecker(filename=str(filename))
@@ -177,16 +199,17 @@ class BugbearTestCase(unittest.TestCase):
         bbc = BugBearChecker(filename=str(filename))
         errors = list(bbc.run())
         expected = self.errors(
-            B014(10, 0, vars=("Exception, TypeError", "", "Exception")),
-            B014(16, 0, vars=("OSError, OSError", " as err", "OSError")),
-            B014(27, 0, vars=("MyError, MyError", "", "MyError")),
-            B014(41, 0, vars=("MyError, BaseException", " as e", "BaseException")),
-            B014(48, 0, vars=("re.error, re.error", "", "re.error")),
+            B014(11, 0, vars=("Exception, TypeError", "", "Exception")),
+            B014(17, 0, vars=("OSError, OSError", " as err", "OSError")),
+            B014(28, 0, vars=("MyError, MyError", "", "MyError")),
+            B014(42, 0, vars=("MyError, BaseException", " as e", "BaseException")),
+            B014(49, 0, vars=("re.error, re.error", "", "re.error")),
             B014(
-                55,
+                56,
                 0,
                 vars=("IOError, EnvironmentError, OSError", "", "OSError"),
             ),
+            B014(74, 0, vars=("ValueError, binascii.Error", "", "ValueError")),
         )
         self.assertEqual(errors, expected)
 
@@ -211,6 +234,82 @@ class BugbearTestCase(unittest.TestCase):
         expected = self.errors(B017(22, 8))
         self.assertEqual(errors, expected)
 
+    def test_b018_functions(self):
+        filename = Path(__file__).absolute().parent / "b018_functions.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+
+        expected = [B018(line, 4) for line in range(16, 26)]
+        expected.append(B018(29, 4))
+        expected.append(B018(32, 4))
+        self.assertEqual(errors, self.errors(*expected))
+
+    def test_b018_classes(self):
+        filename = Path(__file__).absolute().parent / "b018_classes.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+
+        expected = [B018(line, 4) for line in range(17, 27)]
+        expected.append(B018(30, 4))
+        expected.append(B018(33, 4))
+        self.assertEqual(errors, self.errors(*expected))
+
+    def test_b019(self):
+        filename = Path(__file__).absolute().parent / "b019.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+
+        # AST Decorator column location for callable decorators changes in 3.7
+        col = 5 if sys.version_info >= (3, 7) else 4
+        self.assertEqual(
+            errors,
+            self.errors(
+                B019(73, 5),
+                B019(77, 5),
+                B019(81, col),
+                B019(85, col),
+                B019(89, 5),
+                B019(93, 5),
+                B019(97, col),
+                B019(101, col),
+            ),
+        )
+
+    def test_b020(self):
+        filename = Path(__file__).absolute().parent / "b020.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        self.assertEqual(
+            errors,
+            self.errors(
+                B020(8, 4, vars=("items",)),
+                B020(21, 9, vars=("values",)),
+            ),
+        )
+
+    def test_b021_classes(self):
+        filename = Path(__file__).absolute().parent / "b021.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        expected = self.errors(
+            B021(14, 4),
+            B021(22, 4),
+            B021(30, 4),
+            B021(38, 4),
+            B021(46, 4),
+            B021(54, 4),
+            B021(62, 4),
+            B021(70, 4),
+            B021(74, 4),
+        )
+        self.assertEqual(errors, expected)
+
+    def test_b022(self):
+        filename = Path(__file__).absolute().parent / "b022.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        self.assertEqual(errors, self.errors(B022(8, 0)))
+
     def test_b901(self):
         filename = Path(__file__).absolute().parent / "b901.py"
         bbc = BugBearChecker(filename=str(filename))
@@ -272,6 +371,7 @@ class BugbearTestCase(unittest.TestCase):
             B904(10, 8),
             B904(11, 4),
             B904(16, 4),
+            B904(55, 16),
         ]
         self.assertEqual(errors, self.errors(*expected))
 
@@ -279,7 +379,49 @@ class BugbearTestCase(unittest.TestCase):
         filename = Path(__file__).absolute().parent / "b950.py"
         bbc = BugBearChecker(filename=str(filename))
         errors = list(bbc.run())
-        self.assertEqual(errors, self.errors(B950(6, 92, vars=(92, 79))))
+        self.assertEqual(
+            errors,
+            self.errors(
+                B950(7, 92, vars=(92, 79)),
+                B950(12, 103, vars=(103, 79)),
+                B950(14, 103, vars=(103, 79)),
+                B950(21, 97, vars=(97, 79)),
+            ),
+        )
+
+    def test_b9_select(self):
+        filename = Path(__file__).absolute().parent / "b950.py"
+
+        mock_options = Namespace(select=["B950"])
+        bbc = BugBearChecker(filename=str(filename), options=mock_options)
+        errors = list(bbc.run())
+        self.assertEqual(
+            errors,
+            self.errors(
+                B950(7, 92, vars=(92, 79)),
+                B950(12, 103, vars=(103, 79)),
+                B950(14, 103, vars=(103, 79)),
+                B950(21, 97, vars=(97, 79)),
+            ),
+        )
+
+    def test_b9_extend_select(self):
+        filename = Path(__file__).absolute().parent / "b950.py"
+
+        # select is always going to have a value, usually the default codes, but can
+        # also be empty
+        mock_options = Namespace(select=[], extend_select=["B950"])
+        bbc = BugBearChecker(filename=str(filename), options=mock_options)
+        errors = list(bbc.run())
+        self.assertEqual(
+            errors,
+            self.errors(
+                B950(7, 92, vars=(92, 79)),
+                B950(12, 103, vars=(103, 79)),
+                B950(14, 103, vars=(103, 79)),
+                B950(21, 97, vars=(97, 79)),
+            ),
+        )
 
     def test_selfclean_bugbear(self):
         filename = Path(__file__).absolute().parent.parent / "bugbear.py"
